/*******************************************************************************
**	text16 operators
********************************************************************************
**
   Copyright (C)	1997		Jochen Hoenicke
**			1997		Steffen Seeger
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
********************************************************************************
**
**	This code was derived from the following source of information:
**	[1]     Xfree 86 vga driver
**	[2]     Zen of Graphics Programming (Michael Abrash)
*/

#ifndef _S3_TEXT16_INC_
#define _S3_TEXT16_INC_

#if defined(S3_86C964) || defined(S3_86C968) || defined(S3_86C765) || defined(S3_86C325)

	#define	MMIO_ONLY_TEXT
#endif

#if defined(S3_86C928)

	#define	NORMAL_TEXT
#endif

#define TEXTX           (dpy->mode.request.dpp.x)
#define TEXTY           (dpy->mode.request.dpp.y)
#define VIRTUALX        (dpy->fb.size.x)
#define VIRTUALY        (dpy->fb.size.y)
#define VISIBLEX        (dpy->mode.request.visible.x)
#define VISIBLEY        (dpy->mode.request.visible.y)

#ifdef MMIO_ONLY_TEXT

/*
**	MMIO only text mode operations
*/

#ifdef DEBUG_LEVEL
	/*	NOTE: the ASSERT lines should give you a ring if the framebuffer
	**	operations are used with inconsistent parameters. However, we
	**	rely on the OS to catch NULL pointer references!
	*/

	#define	FBSIZE	(fb->size.x*fb->size.y)
#endif


static void s3_text16_move(kgi_pixmap *fb, kgi_uint dst, kgi_uint src, kgi_uint cnt)
{
	register kgiu16 *pbuf = (void *) fb->pbuf;

	ASSERT((src + cnt) <= FBSIZE);
	ASSERT((dst + cnt) <= FBSIZE);
	ASSERT(src != dst);

	dst <<= 2;
	src <<= 2;

	if (dst < src) {

		while (cnt--) {

			mem_out16(mem_in16(pbuf + src), pbuf + dst);
			src += 4;
			dst += 4;
		}

	} else {

		src += (cnt - 1) << 2;
		dst += (cnt - 1) << 2;

		while (cnt--) {

			mem_out16(mem_in16(pbuf + src), pbuf + dst);
			src -= 4;
			dst -= 4;
		}
	}
}

static void s3_text16_set(kgi_pixmap *fb, kgi_uint dst, kgi_pixel val, kgi_uint cnt)
{
	register kgiu16 *pbuf = ((kgiu16 *) fb->pbuf) + (dst << 2);

	ASSERT((dst + cnt) <= FBSIZE);

	while (cnt--) {

		mem_out16(val, pbuf);
		pbuf += 4;
	}
}

static void s3_text16_invert(kgi_pixmap *fb, kgi_uint start, kgi_uint end)
{
	register kgiu16 *pbuf = ((kgiu16 *)fb->pbuf) + (start << 2);
	register kgiu16 *pend = pbuf + (end << 2);

	ASSERT((start < FBSIZE) && (end < FBSIZE));
	ASSERT(start <= end);

	while (pbuf <= pend) {

		register kgiu16 h = mem_in16(pbuf);

		mem_out16((h & 0x88FF) | ((h<<4) & 0x7000) | ((h>>4) & 0x0700),
			pbuf);
		pbuf += 4;
	}
}

static void s3_text16_update(kgi_pixmap *fb, kgi_uint pos, kgi_pixel val)
{
	ASSERT(pos < FBSIZE);
	mem_out16(val, ((kgiu16 *) fb->pbuf) + (pos << 2));
}

static void s3_text16_flush(kgi_pixmap *fb)
{
}

static void s3_text16_copyfrom(kgi_pixmap *fb, void *buf, kgi_uint start, kgi_uint cnt)
{
	register kgiu16 *src = ((kgiu16 *) fb->pbuf) + (start << 2);
	ASSERT((start + cnt) <= FBSIZE);

	while (cnt--) {

		mem_out16(*src, ((kgiu16 *) buf)++);
		src += 4;
	}
}

static void s3_text16_copyto(kgi_pixmap *fb, void *buf, kgi_uint start, kgi_uint cnt)
{
	register kgiu16 *dst = ((kgiu16 *) fb->pbuf) + (start << 2);
	ASSERT((start + cnt) <= FBSIZE);

	while (cnt--) {

		mem_out16(*(((kgiu16 *) buf)++), dst);
		dst += 4;
	}
}

/*	color mode attribute to pixel value mapping. We care of the foreground
**	and background colors. The rest is just best effort.
*/
static kgi_pixel s3_text16_atop(kgi_pixmap *fb, kgi_attr attr)
{
	register kgi_pixel
	result = (attr & ATTR_REVERSE)
			? (((attr & 0x0007) << 8) | ((attr & 0x0700) << 4))
			: (((attr & 0x0007) << 12) | (attr & 0x0700));

	if (attr & ATTR_BLINK) {

		result |= 0x8000;
	}

	if (attr & ATTR_BOLD) {

		result |= 0x0800;
	}

	return result;
}

static kgi_attr s3_text16_ptoa(kgi_pixmap *fb, kgi_pixel val)
{
	return	  ((val & 0x8000) ? ATTR_BLINK : 0)
		| ((val & 0x0800) ? ATTR_BOLD  : 0)
		| ((val & 0x7000) >> 12) | (val & 0x0F00);
}

/*	font postion to pixel mapping and vice versa. This cares about 256 
**	character fonts only (at least for now).
*/
static kgi_pixel s3_text16_ftop(kgi_pixmap *fb, kgi_uint pos)
{
	return pos & 0x00FF;
}

static kgi_uint s3_text16_ptof(kgi_pixmap *fb, kgi_pixel val)
{
	return val & 0x00FF;
}

static struct kgi_fb_ops s3_text16 = 
{
	s3_text16_move,
	s3_text16_set,
	s3_text16_invert,
	s3_text16_update,
	s3_text16_flush,
	s3_text16_copyfrom,
	s3_text16_copyto,

	s3_text16_atop,
	s3_text16_ptoa,
	s3_text16_ftop,
	s3_text16_ptof
};

/*
**	MMIO only text mode font setting
*/

static void do_set_font(struct kgi_display *dpy, int no, kgi_uint first,
	kgi_uint last, struct kgi_font *font)
{
	register kgiu32 *fbuf = (void *) dpy->fb.pbuf;
	register kgiu8 *data = font->data + font->dy*first;

	if (font->data) {

		while (first <= last) {

			kgi_uint ggiindex = (first & 0xFF) + (no << 13);
			int i;

			for (i = 0; i < font->dy; i++, ggiindex += 0x100) {

				mem_out32((mem_in32(fbuf+ggiindex) & 0xFF00FFFF) |
					(data[i] << 16), fbuf+ggiindex);
			}
			for ( ; i < 32; i++, ggiindex += 0x100) {

				mem_out32(mem_in32(fbuf + ggiindex) & 0xFF00FFFF,
					fbuf + ggiindex);
			}

			first++;
			data += font->dy;
		}

        } else {

		for ( ; first <= last; first++) {

			kgi_uint ggiindex = first + (no << 13) + 32*0x100;
			int i;

			for (i = 33; --i; ggiindex -= 0x100) {

				mem_out32(mem_in32(fbuf + i) & 0xFF00FFFF,
					fbuf + i);
			}
		}
	}
}

#else	/* #ifdef MMIO_ONLY_TEXT */

static void do_set_font(struct kgi_display *dpy, int no, kgi_uint first,
	kgi_uint last, struct kgi_font *font)
{
	IS_CRITICAL;
	register GET_PRIVATE;
	kgiu8 *data, *fbuf = MEM_ACCEL;
	kgiu8 SR2, SR4, GR4, GR5, GR6, CR3A;
	register int i,j;

	if (font == NULL) {

		return;
	}
	fbuf += (no * 0x2000) + (first << 5);
	data = font->data + font->dy * first;

	/*	prepare hardware to access font plane
	*/
	ENTER_CRITICAL;

		/*	save register state
		*/
		SR2  = SEQ_in8(dpy, 0x02);
		SR4  = SEQ_in8(dpy, 0x04);
		GR4  = GRC_in8(dpy, 0x04);
		GR5  = GRC_in8(dpy, 0x05);
		GR6  = GRC_in8(dpy, 0x06);
		CR3A = CRT_in8(dpy, 0x3A);

		SEQ_out8(dpy, SR02_PLANE2, 0x02);    /* write to plane 2  */
		SEQ_out8(dpy, SR04_256K_ACCESS | SR04_NO_ODDEVEN, 0x04);

		GRC_out8(dpy, 2, 0x04);	       /* read from plane 2 */
		GRC_out8(dpy, GR5_WRITEMODE0, 0x05);
		GRC_out8(dpy, GR6_MAP_A0_64 | GR6_GRAPHMODE, 0x06);

		CRT_out8(dpy, CR3A | CR3A_HISPEEDTXT, 0x3A);
	
	LEAVE_CRITICAL;

	/* NOTE	no printk()'s are allowed in here unless klogd is active !
	*/

	if (data) {

		/*	set font if data is valid
		*/
		register int cnt = 32 - font->dy;

		for (i = first; i <= last; i++) {

			for (j = font->dy; j--; ) {

				mem_out8(*(data++), fbuf++);
			}
			for (j = cnt; j--; ) {

				mem_out8(0, fbuf++);
			}
		}

	} else {

		register kgiu32 *h = (void *) fbuf;

		/*	clear font
		*/
		for (i = first; i <= last; i++ ) {

			mem_out32(0, h++);
			mem_out32(0, h++);
			mem_out32(0, h++);
			mem_out32(0, h++);

			mem_out32(0, h++);
			mem_out32(0, h++);
			mem_out32(0, h++);
			mem_out32(0, h++);
		}
	}

	/*	restore hardware state
	*/
	ENTER_CRITICAL;

		CRT_out8(dpy, (CRT_in8(dpy, 0x0A) & ~CR0A_CURSOR_START_MASK) |
			((font->base + 1) & CR0A_CURSOR_START_MASK), 0x0A);
		CRT_out8(dpy, (CRT_in8(dpy, 0x0B) & ~CR0B_CURSOR_END_MASK) |
			((font->base + 2) & CR0B_CURSOR_END_MASK), 0x0B);

		SEQ_out8(dpy, SR2,  0x02);
		SEQ_out8(dpy, SR4,  0x04);
		GRC_out8(dpy, GR4,  0x04);
		GRC_out8(dpy, GR5,  0x05);
		GRC_out8(dpy, GR6,  0x06);
		CRT_out8(dpy, CR3A, 0x3A);

	LEAVE_CRITICAL;
}
#endif

static void s3_text_set_font(struct kgi_display *dpy, struct kgi_font *font,
			  kgi_uint from, kgi_uint to)
{
	do_set_font(dpy, 0, from, to, font);
	do_set_font(dpy, 1, from, to, font);
}

static void s3_text_set_origin(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	IS_CRITICAL;
	GET_PRIVATE;

	register kgi_uint xpix = x / TEXTX;
	register kgi_uint xdot = x % TEXTX;
	register kgi_uint ypix = y / TEXTY;
	register kgi_uint ydot = y % TEXTY;

	register kgi_uint offset = xpix + ypix*dpy->fb.size.x;
	private->origin = private->frame + offset;

	ENTER_CRITICAL;

		if (private->org_offs != offset) {

			CRT_out8(dpy, offset >> 8, CRT_STARTADDR_H);
			CRT_out8(dpy, offset,      CRT_STARTADDR_L);
			private->org_offs = offset;
		}

		if (private->org_ydot != ydot) {

			CRT_out8(dpy, ydot & CR08_ROWSCAN_MASK, CRT_PRESETROW);
			private->org_ydot = ydot;
		}

		if (private->org_xdot != xdot) {

			wait_display_off(dpy);

			if (dpy->mode.request.dpp.x == 9) {

				ATC_out8(dpy, (xdot == 9) ? 0x08 :
					(xdot-1) & AR13_DATA_MASK, ATC_HPAN);
			} else {

				ATC_out8(dpy, xdot & AR13_DATA_MASK, ATC_HPAN);
			}
			private->org_xdot = xdot;
		}

	LEAVE_CRITICAL;
}


static void s3_set_split(struct kgi_display *dpy, int line)
{
	GET_PRIVATE;
	register kgiu8 foo;

	if (private->split == line) {

		return;	
	}
	private->split = line;

	CRT_out8(dpy, line & 0xFF, CRT_LINECOMPARE);

	foo = (line & 0x0100)
		? CRT_in8(dpy, 0x07) |  CR07_LINECOMPARE_B8
		: CRT_in8(dpy, 0x07) & ~CR07_LINECOMPARE_B8;
	CRT_out8(dpy, foo, 0x07);

	foo = (line & 0x0200)
		? CRT_in8(dpy, 0x09) |  CR09_LINECOMPARE_B9
		: CRT_in8(dpy, 0x09) & ~CR09_LINECOMPARE_B9;
	CRT_out8(dpy, foo, 0x09);

	foo = (line & 0x0400)
		? CRT_in8(dpy, 0x5E) |  CR5E_LINECOMPARE_B10
		: CRT_in8(dpy, 0x5E) & ~CR5E_LINECOMPARE_B10;
	CRT_out8(dpy, foo, 0x5E);
}

#define s3_set_frame NULL

static void s3_set_origin(struct kgi_display *dpy, kgi_uint dx, kgi_uint dy)
{
	int x = dx + dy*dpy->mode.request.virt.x;

	switch(dpy->mode.request.graphtype) {
	case KGIGT_4BIT:	x >>= 1; break;
	case KGIGT_15BIT:
	case KGIGT_16BIT:	x <<= 1; break;
	case KGIGT_32BIT:	x <<= 2; break;
	default:	break;
	}

#if defined(S3_86C765) || defined(S3_86C964) || defined(S3_86C968) || defined(S3_86C325)

	CRT_out8(dpy, (x >>  2) & 0xff, CRT_STARTADDR_L);    /*  sa2-sa9 */
	CRT_out8(dpy, (x >> 10) & 0xff, CRT_STARTADDR_H);    /* sa10-sa17 */
	CRT_out8(dpy, (x >> 18) & 0x0f, CRT_EXTCONTROL3 );

#elif defined(S3_86C928)

	x >>= 2;
	CRT_out8(dpy, x         & 0xff, CRT_STARTADDR_L);	/*  sa2-sa9  */
	CRT_out8(dpy, (x >>  8) & 0xff, CRT_STARTADDR_H);	/* sa10-sa17 */
	CRT_out8(dpy, (CRT_in8(dpy, CRT_MEMCONFIG0) & ~0x30) |
		((x >> 12) & 0x30), CRT_MEMCONFIG0);		/* sa18-sa19 */
	CRT_out8(dpy, (CRT_in8(dpy, CRT_SYSCONTROL2) & ~0x03) |
		((x >> 18) & 0x03), CRT_SYSCONTROL2);		/* sa20-sa21 */
#else
	#error no s3_set_origin, check this code!
#endif
}

/*
**	Cursor handling
*/

/*	hardware cursor
*/
static void s3_show_hw_text_cursor(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	register GET_PRIVATE;
	register kgi_uint pos = private->origin + 
			(x / TEXTX) + (y / TEXTY)*dpy->fb.size.x;

	pos = (pos < private->fb_size) ? pos : pos - private->fb_size;
	ASSERT(pos < private->fb_size);

	DEBUG2("cursor (%i,%i) = %i", x, y, pos);
	CRT_out8(dpy, pos >> 8, CRT_CURSORADDR_H);
	CRT_out8(dpy, pos,      CRT_CURSORADDR_L);
}

static void s3_hide_hw_text_cursor(struct kgi_display *dpy)
{
	CRT_out8(dpy, 0xFF, CRT_CURSORADDR_H);
	CRT_out8(dpy, 0xFF, CRT_CURSORADDR_L);
}

/*	software pointer and cursor
*/
#define	PTRAND	(private->ptr_and)
#define	PTRXOR	(private->ptr_xor)
#define	PTROLD	(private->ptr_old)

#define	CURAND	(private->cur_and)
#define	CURXOR	(private->cur_xor)
#define	CUROLD	(private->cur_old)

static void s3_show_sw_text_pointer(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	register GET_PRIVATE;
	register kgiu32 old = PTROLD;
	register kgiu32 cur = CUROLD;
	register kgiu32 new = private->origin +
		(x / TEXTX)  +  (y / TEXTY) * VIRTUALX;
	register kgiu16 *p = (kgiu16 *) dpy->fb.pbuf;

	ASSERT((x < VISIBLEX) && (y < VISIBLEY));

	new = (new < private->fb_size) ? new : new - private->fb_size;
	ASSERT(new < private->fb_size);

	if (old & 0x8000) {

		#if defined(MMIO_ONLY_TEXT)

			mem_out16(old >> 16, p + ((old & 0x7FFF) << 2));

		#elif defined(NORMAL_TEXT)

			mem_out16(old >> 16, p + (old & 0x7FFF));
		#else
			#error implement software text mode pointer!
		#endif
	}

	#if defined(NORMAL_TEXT)
		p += new;
	#elif defined(MMIO_ONLY_TEXT)
		p += new << 2;
	#else
		#error implement software textmode pointer!
	#endif

	if ((cur & 0x8000) && ((cur & 0x7FFF) == (new & 0x7FFF))) {

		new |= cur & 0xFFFF0000;

	} else {

		new |= mem_in16(p) << 16;
	}

	PTROLD = new | 0x8000;
	mem_out16((mem_in16(p) & PTRAND) ^ PTRXOR, p);
}

static void s3_hide_sw_text_pointer(struct kgi_display *dpy)
{
	GET_PRIVATE;
	register kgiu32 old = PTROLD;
	register kgiu16 *p = (kgiu16 *) dpy->fb.pbuf;
	ASSERT(p);

	if (old & 0x8000) {

		#if defined(MMIO_ONLY_TEXT)

			mem_out16(old >> 16, p + ((old & 0x7FFF) << 2));

		#elif defined(NORMAL_TEXT)

			mem_out16(old >> 16, p + (old & 0x7FFF));

		#else
			#error implement sw_pointer_hide()
		#endif
	}

	PTROLD = 0;
}

#define	s3_undo_sw_text_pointer	s3_hide_sw_text_pointer

static void s3_show_sw_text_cursor(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	GET_PRIVATE;
	register kgiu32 old = CUROLD;
	register kgiu32 cur = PTROLD;
	register kgiu32 new = private->origin +
		(x / TEXTX)  +  (y / TEXTY) * VIRTUALX;
	register kgiu16 *p = (kgiu16 *) dpy->fb.pbuf;

	ASSERT((x < VISIBLEX) && (y < VISIBLEY));

	new = (new < private->fb_size) ? new : new - private->fb_size;
	ASSERT(new < private->fb_size);

	if (old & 0x8000) {

		#if defined(MMIO_ONLY_TEXT)

			mem_out16(old >> 16, p + ((old & 0x7FFF) << 2));

		#elif defined(NORMAL_TEXT)

			mem_out16(old >> 16, p + (old & 0x7FFF));
		#else
			#error implement software text mode pointer!
		#endif
	}

	#if defined(NORMAL_TEXT)
		p += new;
	#elif defined(MMIO_ONLY_TEXT)
		p += new << 2;
	#else
		#error implement software textmode pointer!
	#endif

	if ((cur & 0x8000) && ((cur & 0x7FFF) == (new & 0x7FFF))) {

		new |= cur & 0xFFFF0000;

	} else {

		new |= mem_in16(p) << 16;
	}

	CUROLD = new | 0x8000;
	mem_out16((mem_in16(p) & CURAND) ^ CURXOR, p);
}

static void s3_hide_sw_text_cursor(struct kgi_display *dpy)
{
	GET_PRIVATE;
	register kgiu32 old = CUROLD;
	register kgiu16 *p = (kgiu16 *) dpy->fb.pbuf;
	ASSERT(p);

	if (old & 0x8000) {

		#if defined(MMIO_ONLY_TEXT)

			mem_out16(old >> 16, p + ((old & 0x7FFF) << 2));

		#elif defined(NORMAL_TEXT)

			mem_out16(old >> 16, p + (old & 0x7FFF));

		#else
			#error implement sw_pointer_hide()
		#endif
	}

	CUROLD = 0;
}

#define	s3_undo_sw_text_cursor	s3_hide_sw_text_cursor

#undef	PTRAND
#undef	PTRXOR
#undef	PTROLD

#undef	CURAND
#undef	CURXOR
#undef	CUROLD

#undef	TEXTX
#undef	TEXTY
#undef	VIRTUALX
#undef	VISIBLEX
#undef	VISIBLEY

#endif /* _S3_TEXT16_INC */
