/*******************************************************************************
**	Chips & Technologies 655xx ramdac driver
********************************************************************************
**
   Copyright (C) 1997	Stephan Meyer
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
********************************************************************************
**
**	$Log: 655xx.c,v $
**	Revision 1.8  1998/11/19 02:43:04  marcus
**	License change
**	
**	Revision 1.7  1998/10/25 20:03:25  brian
**	Changed cmd to *cmd in ramdac_check_mode and accel_check_mode, plus
**	spot-cleaned a few places Marcus missed during namespace cleanup.
**	(Will get chipset/clock stuff soon, working around Marcus.)
**	
**	Revision 1.6  1998/10/25 18:04:11  marcus
**	More namespace cleanups
**	
**	Revision 1.5  1998/10/25 02:05:42  marcus
**	Namespace cleanup to make KGI/LibGGI interaction possible
**	
**	Revision 1.4  1998/10/04 03:16:40  brian
**	Added private data hook to KGI_COMMON -- clean and recompile kgicon!
**	
**	Revision 1.3  1998/08/25 12:12:08  marcus
**	*** empty log message ***
**	
**	Revision 1.2  1998/08/19 20:47:47  marcus
**	Changed ggi/*.h includes to kgi/*.h includes
**	Fixed most of the remaining graphics -> accel changes
**	
**	Revision 1.1  1998/07/07 07:51:01  taylorj
**	*** empty log message ***
**	
**	Revision 1.4  1997/12/18 23:51:24  becka
**	More ramdacs set right ...
**	
**	Revision 1.3  1997/12/14 21:56:17  becka
**	uint/sint namespace fixed ... please test ...
**	
**	Revision 1.2  1997/12/03 15:47:09  seeger
**	small bug in header includes fixed.
**	
**	Revision 1.1  1997/10/22 17:35:46  seeger
**	Initial C&T driver checkin
**	
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Stephan_Meyer
#define	DRIVER_NAME	"Chips 655xx ramdac driver"
#define	DRIVER_REV	"$Revision: 1.8 $"
#define	DEBUG_LEVEL	0

#include <kgi/module.h>

#include "Chips/chips.h"

#define DAC_PRECISION   6

struct kgi_ramdac ramdac = 
{
	"Chips & Technologies",		/* Manufacturer		*/
	"655xx",			/* Model		*/
	GM_TEXT16 | GM_8BIT |
	GM_16BIT | GM_24BIT,		/* Supported modes	*/
	0,			 	/* Flags		*/
	{ 0xFFFF, 0xFFFF },		/* Maximum resolution	*/
	NULL,				/* Private data hook	*/
	DAC_8BIT_PB,  			/* Device type		*/
	{       0,  32000000 },		/* Normal mode		*/
	{       0,         0 },		/* 2xclk mode		*/
	{       /* autodetect */
		{	  0,  25175000 },	/* KGIGT_TEXT16	*/
		{	  0,         0 },	/* KGIGT_TEXT32	*/
		{         0,         0 },	/* KGIGT_1BIT	*/
		{	  0,         0 },	/* KGIGT_4BIT	*/
		{         0,         0 },	/* KGIGT_8BIT	*/	
		{	  0,         0 },	/* KGIGT_15BIT	*/
		{ 	  0,         0 },	/* KGIGT_16BIT	*/
		{         0,         0 },	/* KGIGT_24BIT	*/
		{	  0,         0 }	/* KGIGT_32BIT	*/
	}
};

extern int	chips_model;
extern ChipsRegs chips_resetregs;

/* taken from the cirrus ramdac driver */

static void dac_set_clut(struct kgi_display *dpy, kgi_clut *clut,
			 kgi_uint first, kgi_uint cnt) {
	kgiu16 *data=(kgiu16 *)(clut->data+first);
	cnt*=3;
	outb(first,0x3C8);
	while (cnt--) outb(*(data++)>>10,0x3C9);
}

static void dac_get_clut(struct kgi_display *dpy, kgi_clut *clut,
			 kgi_uint first, kgi_uint cnt) {
	kgiu16 *data=(kgiu16 *)(clut->data+first);
	cnt*=3;
	outb(first,0x3C7);
	while (cnt--) *(data++)=inb(0x3C9)*1040;
}

/* GGI interface */

int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME DRIVER_REV);
	switch (chips_model) {
		case CHIPS_65545 :
			if (chips_resetregs.extension[0x6C] & 2) {
				ramdac.clock[KGIGT_8BIT].max=68000000;
				ramdac.clock[KGIGT_16BIT].max=34000000;
				ramdac.clock[KGIGT_24BIT].max=22667000;
			} else {
				ramdac.clock[KGIGT_8BIT].max=56000000;
				ramdac.clock[KGIGT_16BIT].max=28322000;
				ramdac.clock[KGIGT_24BIT].max=18667000;
			}
			break;
		case CHIPS_65548 :
			ramdac.clock[KGIGT_8BIT].max=80000000;
			ramdac.clock[KGIGT_16BIT].max=40000000;
			ramdac.clock[KGIGT_24BIT].max=26667000;
			break;
		case CHIPS_65550 :
		case CHIPS_65554 :
			io_out8(0x0A,0x3D0);
			if (io_in8(0x3D1) & 2) {
				ramdac.clock[KGIGT_8BIT].max=110000000;
				ramdac.clock[KGIGT_16BIT].max=110000000;
				ramdac.clock[KGIGT_24BIT].max=110000000;
			} else {
				ramdac.clock[KGIGT_8BIT].max=80000000;
				ramdac.clock[KGIGT_16BIT].max=80000000;
				ramdac.clock[KGIGT_24BIT].max=80000000;
			}
			break;
	}

	card->ramdac=&ramdac;

	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed.");
}

int kgim_ramdac_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd *cmd)
{
	kgi_uint newclock = mode->dclk;
  
	switch (mode->request.graphtype) {

	case KGIGT_TEXT16:
		break;

	case KGIGT_8BIT:
		break;

	case KGIGT_16BIT:
		break;

	case KGIGT_24BIT:
		break;
  
	default:
  		DEBUG2("couldn't handle graphtype %i", 
		       mode->request.graphtype);
		return -E(RAMDAC,MODE_CHECK);
	}

	switch (*cmd) {

	case CMD_PROPOSE:
		mode->lclk.mul = mode->clk.mul = 1;
		mode->lclk.div = mode->clk.div = 1;
		return EOK;

	case CMD_LOWER:

		if ((newclock < ramdac.clock[mode->request.graphtype].min) ||
			(newclock < ramdac.normal.min)) {

			DEBUG2("DCLK too low (%i Hz).", newclock);
			return -E(RAMDAC,CLK_LOW_LIMIT);
		}

		if (ramdac.clock[mode->request.graphtype].max < newclock) {

			newclock = ramdac.clock[mode->request.graphtype].max;
		}
		mode->dclk = newclock;
		return EOK;

	case CMD_RAISE:
		if ((newclock > ramdac.clock[mode->request.graphtype].max) ||
			(newclock > ramdac.normal.max)) {

			DEBUG2("DCLK too high (%i Hz).", newclock);
			return -E(RAMDAC,CLK_HIGH_LIMIT);
		}

		if (newclock < ramdac.clock[mode->request.graphtype].min) {

			newclock = ramdac.clock[mode->request.graphtype].min;
		}

		if (newclock < ramdac.normal.min) {

			newclock = ramdac.normal.min;
		}

		mode->dclk = newclock;
		return EOK;

     case CMD_CHECK:
		if ((mode->lclk.mul != 1) || (mode->lclk.div != 1) ||
		    (mode->clk.mul != 1) || (mode->clk.div != 1)) {

			return -E(RAMDAC,MODE_CHECK);
		}

		if ((mode->dclk < ramdac.clock[mode->request.graphtype].min) ||
		    (mode->dclk < ramdac.normal.min)) {

			return -E(RAMDAC,CLK_LOW_LIMIT);
		}

		if ((mode->dclk > ramdac.clock[mode->request.graphtype].max) ||
		    (mode->dclk > ramdac.normal.max)) {
			
			return -E(RAMDAC,CLK_HIGH_LIMIT);
		}

		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(RAMDAC,BOGUS);
	}
}

void kgim_ramdac_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	switch (mode->request.graphtype) {

	case KGIGT_TEXT16:
	case KGIGT_8BIT:
	case KGIGT_16BIT:
	case KGIGT_24BIT:
		break;

	default:
		INTERNAL_ERROR;
	}
	dpy->set_clut = dac_set_clut;
}

int kgim_ramdac_command(struct kgi_graphic *graph, kgi_uint cmd, kgi_uint arg)
{
	/* This may set 
	 * -  Hardware cursor
	 * -  CLUT (or gamma correction table)
	 */
	kgi_clut clut;

	switch (cmd) { 

	case RAMDAC_SETCLUT:
		clut.size = 256;
		clut.data = (kgi_color *)(graph->io_buf);
		dac_set_clut(graph->dev.dpy, &clut, 0, 256);
		return 0;

	case RAMDAC_GETCLUT:
		clut.size = 256;
		clut.data = (kgi_color *)(graph->io_buf);
		graph->io_res = graph->io_buf;
		dac_get_clut(graph->dev.dpy, &clut, 0, 256);
		return 0;
	}
  	return -E(RAMDAC, NOSUP_ALWAYS_CANT);  /* The ramdac can't do this */
}

